/*
** AUTHORS:
** Rama Alebouyeh (rama[at]cs.ucsb.edu)
** Matthew S Allen (msa[at]cs.ucsb.edu)
*/

#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <stdlib.h>
#include "route.h"
#include "host.h"
#include "key.h"
#include "jrb.h"
#include "jval.h"
#include "log.h"


#define LEAFSET_SIZE 8   /* (must be even) excluding node itself */
#define MAX_ROW KEY_SIZE/BASE_B
#define MAX_COL power(2,BASE_B)

typedef struct {
  ChimeraHost *me;
  char *keystr;
  ChimeraHost ***table;
  ChimeraHost **leftleafset;
  ChimeraHost **rightleafset;
  pthread_mutex_t lock;
} RouteGlobal;

void leafset_update(ChimeraState *state,ChimeraHost *host, int joined, ChimeraHost **deleted, ChimeraHost **added);
int leafset_size(ChimeraHost **arr);
void leafset_insert(ChimeraState *state,ChimeraHost *host ,int right_or_left, ChimeraHost **deleted, ChimeraHost **added); 
void leafset_delete(ChimeraState *state,ChimeraHost *host, int right_or_left, ChimeraHost **deleted);
void leafset_print(ChimeraState *state);
void sort_hosts (void * logs, ChimeraHost **hosts,Key key,int size);
int power ( int base, int n);
int hexalpha_to_int(int c);

void route_keyupdate(void *routeglob,ChimeraHost *me) {
  
  RouteGlobal *rg = (RouteGlobal *) routeglob;
  rg->me = me;
  sprintf(rg->keystr,"%s", rg->me->key.keystr);
}


/** route_init:
** Ininitiates routing table and leafsets 
*/
void * route_init(ChimeraHost * me) {

  int i,j;
  RouteGlobal *rg ;
  rg = (RouteGlobal *)malloc(sizeof(RouteGlobal));

  /* allocate memory for routing table */
  rg->table=(ChimeraHost ***)malloc(sizeof (struct ChimeraHost **) * MAX_ROW);
  for ( i=0; i< MAX_ROW; i++){
     rg->table[i]=(ChimeraHost **)malloc(sizeof(ChimeraHost *)*MAX_COL);
     for(j=0;j<MAX_COL;j++)
       rg->table[i][j]=NULL;
     
  }
 
  rg->keystr = (char *) malloc(sizeof(char)*MAX_ROW);
  route_keyupdate((void *)rg, me);
  
  
  /* allocate memory for leafsets */
  rg->leftleafset=(ChimeraHost **)malloc(sizeof(ChimeraHost *)*(LEAFSET_SIZE/2)+1);
  rg->rightleafset=(ChimeraHost **)malloc(sizeof(ChimeraHost *)*(LEAFSET_SIZE/2)+1);
  for(i=0;i<(LEAFSET_SIZE/2)+1;i++){
    rg->leftleafset[i]=NULL;
    rg->rightleafset[i]=NULL;
  }
  
  pthread_mutex_init(&rg->lock, NULL);

  return ((void*)rg);
}

ChimeraHost **route_get_table(ChimeraState *state){
  ChimeraHost **ret;
  int i,j,k,count;
  RouteGlobal *routeglob = (RouteGlobal *) state->route;

  pthread_mutex_lock(&routeglob->lock);
  
  count = 0;
  
  for(i=0;i<MAX_ROW;i++)
    for(j=0;j<MAX_COL;j++)
      if(routeglob->table[i][j]!=NULL)
	count++;
  ret = (ChimeraHost **) malloc(sizeof(ChimeraHost *) * (count + 1));
  k=0;
  for(i=0;i<MAX_ROW;i++)
    for(j=0;j<MAX_COL;j++)
      if(routeglob->table[i][j]!=NULL)
	ret[k++]= host_get(state, routeglob->table[i][j]->name, routeglob->table[i][j]->port);

  ret[k]=NULL;
  pthread_mutex_unlock(&routeglob->lock);
  return ret;

}
/** route_row_lookup:key 
** return the row in the routing table that matches the longest prefix with #key# 
*/
ChimeraHost **route_row_lookup(ChimeraState *state,Key key){
  ChimeraHost **ret;
  int i,j,k,count;
  RouteGlobal *routeglob = (RouteGlobal *) state->route;

  pthread_mutex_lock(&routeglob->lock);

  
  i= key_index(state->log, routeglob->me->key,key);

  /* find out the number of hosts exists in the matching row */
  count = 0;
  for(j=0;j<MAX_COL;j++)
    if(routeglob->table[i][j]!=NULL)
      count++;

  ret = (ChimeraHost **) malloc(sizeof(ChimeraHost *) * (count + 2));
  k=0;
  for(j=0;j<MAX_COL;j++)
    if(routeglob->table[i][j]!=NULL)
      ret[k++]= host_get(state, routeglob->table[i][j]->name, routeglob->table[i][j]->port);

  ret[k++]=host_get(state, routeglob->me->name,routeglob->me->port);
  ret[k]=NULL;

  pthread_mutex_unlock(&routeglob->lock);

  return ret;
}
/** route_lookup:
* returns an array of #count# nodes that are acceptable next hops for a
* message being routed to #key#. #is_save# is ignored for now.
*/

ChimeraHost **route_lookup(ChimeraState *state,Key key, int count, int is_safe){
  int i,j;
  ChimeraHost *leaf, *tmp;
  int Lsize,Rsize;
  Key Lrange,Rrange,dif1,dif2;
  ChimeraHost **ret;
  Key diffs[LEAFSET_SIZE+1+(MAX_ROW * MAX_COL)];
  int index=0;
  int size = 0;
  ChimeraHost **hosts;
  RouteGlobal *routeglob = (RouteGlobal *) state->route;

  pthread_mutex_lock(&routeglob->lock);
  
  log_message(state->log, LOG_ROUTING, "%s: lookup for key %s!\n",routeglob->me->key.keystr,key.keystr);
  

  /* brute force method to solve count requirements */

  /*calculate the leafset and table size*/
  Lsize=leafset_size(routeglob->leftleafset);
  Rsize=leafset_size(routeglob->rightleafset);


  /* confirm the leafset range */
  if(Rsize==0)
    key_assign(&Rrange, routeglob->me->key);
  else
    key_assign(&Rrange, routeglob->rightleafset[Rsize-1]->key);

  if(Lsize==0)
    key_assign(&Lrange, routeglob->me->key);
  else   
    key_assign(&Lrange,routeglob->leftleafset[Lsize-1]->key);
    
  hosts =  (ChimeraHost **) malloc(sizeof(ChimeraHost *) * (LEAFSET_SIZE+1+(MAX_ROW * MAX_COL)));

  
  leaf = routeglob->me;
  log_message(state->log, LOG_ROUTING, " me: (%s, %d, %s)\n", leaf->name, leaf->port, leaf->key.keystr);
  
  hosts[index++]= routeglob->me;
 
  
  /* look left */
  for(i = 0; i < Lsize; i++) {
    leaf = routeglob->leftleafset[i];
    log_message(state->log, LOG_ROUTING, " Left[%d] (%s, %d, %s): ",i, leaf->name, leaf->port, leaf->key.keystr);
    hosts[index++] = leaf;
  }
  
  /* look right */
  for(i = 0; i < Rsize; i++) {
    leaf = routeglob->rightleafset[i];
    log_message(state->log, LOG_ROUTING, " Right[%d] (%s, %d, %s)\n",i, leaf->name, leaf->port, leaf->key.keystr);
    hosts[index++] =leaf; 
  }
  
  i=key_index(state->log, routeglob->me->key,key);
    
  for(j=0;j<MAX_COL;j++)
    if(routeglob->table[i][j]!=NULL) {
      leaf = routeglob->table[i][j];
      log_message(state->log, LOG_ROUTING, " Table[%d][%d] (%s, %d, %s)\n",i,j,leaf->name, leaf->port, leaf->key.keystr);
      hosts[index++] =leaf; 
    }
  hosts[index]=NULL;
  sort_hosts (state->log, hosts, key,index);
  
  /* find the best #count# entries that we looked at... could be much better */
  ret = (ChimeraHost **) malloc(sizeof(ChimeraHost *) * (count + 1));
  for(i = 0, j = 0; (i - j) < count && hosts[i]!=NULL; i++) {
    tmp = hosts[i];
    
    if((i - j) > 0 && key_equal(ret[(i-j)-1]->key ,tmp->key)) {
      j++;
      continue;
    }
    log_message(state->log, LOG_ROUTING, " Result[%d] (%s, %d, %s)\n",i, tmp->name, tmp->port, tmp->key.keystr);
    ret[i-j] = host_get(state, tmp->name, tmp->port);
  }
  ret[i-j] = NULL;

  // to prevent bouncig
  key_distance(state->log,&dif1,key,ret[0]->key);
  key_distance(state->log,&dif2,key,routeglob->me->key);  
  if(count ==1 && key_equal(dif1, dif2))
    ret[0]=routeglob->me;
  
    
  pthread_mutex_unlock(&routeglob->lock);

  
  if(ret[0]!=NULL)
    log_message(state->log, LOG_ROUTING, "RESULT: %s\n",ret[0]->key.keystr);
  else 
    log_message(state->log, LOG_ROUTING, "RESULT: %s\n",routeglob->me->key.keystr);

  return(ret);
}

void sort_hosts (void * logs, ChimeraHost **hosts,Key key,int size){
  int  i,j;
  ChimeraHost *tmp;
  Key dif1;
  Key dif2;
  
  for(i=0;i<size;i++){
    for(j=i+1;j<size;j++){
      if(hosts[i]!=NULL && hosts[j]!=NULL){
	key_distance(logs, &dif1,hosts[i]->key,key);
	key_distance(logs, &dif2,hosts[j]->key,key);	
	if(key_comp(dif2,dif1)<0){		
	  tmp = hosts[i];
	  hosts[i]=hosts[j];
	  hosts[j]=tmp;
	}
      }
      
    }
  }
  
}

int leafset_size(ChimeraHost **arr){
  int i=0;
  for ( i=0;arr[i]!=NULL;i++);
  return i;

 
}

void leafset_update(ChimeraState *state,ChimeraHost *host, int joined, ChimeraHost **deleted, ChimeraHost **added){

  int Lsize=0;
  int Rsize=0;
  Key midpoint;
  RouteGlobal *routeglob = (RouteGlobal *) state->route;

  Lsize=leafset_size(routeglob->leftleafset);
  Rsize=leafset_size(routeglob->rightleafset);

  key_midpoint(state->log, &midpoint,routeglob->me->key);
  
  if(joined)
    {
      /* key falls in the right side of node */
      if( key_between (state->log, host->key,routeglob->me->key, midpoint)){
	if(Rsize<LEAFSET_SIZE/2 || key_between(state->log, host->key,routeglob->me->key,routeglob->rightleafset[Rsize-1]->key)){
	  /* insert in Right leafset */
	  leafset_insert(state, host,1,deleted,added); 
	}
      }
      
      /* key falls in the left side of the node*/
      else 
	if(Lsize<LEAFSET_SIZE/2 || key_between(state->log, host->key,routeglob->leftleafset[Lsize-1]->key,routeglob->me->key ))
	  /* insert in Left leafset */
	  leafset_insert(state,host,0,deleted,added);
    }
  else
    {
      if( key_between (state->log, host->key,routeglob->me->key, midpoint)){
	leafset_delete(state,host,1,deleted);
      }
      else 
	leafset_delete(state,host,0,deleted);
    }

}

extern void chimera_update_upcall(ChimeraState *state, Key *k, ChimeraHost *h, int joined);

void leafset_delete(ChimeraState *state,ChimeraHost *host, int right_or_left, ChimeraHost **deleted){
  int i=0 ,size;
  int match=0;
  ChimeraHost **p;
  RouteGlobal *routeglob = (RouteGlobal *) state->route;

  /*inert in right leafset*/
  if(right_or_left==1){
    size=leafset_size(routeglob->rightleafset);
    p=routeglob->rightleafset;
  }
  /*insert in left leafset*/
  else{
    size=leafset_size(routeglob->leftleafset);
    p=routeglob->leftleafset;
  }


  for(i=0;i<size && !(key_equal(p[i]->key,host->key)) ;i++);
  if(i<size){
    *deleted = p[i];
    match=1;
  }

  if(match){
    do{
      p[i]=p[i+1];
      i++;
    } while(i<size-1);
    p[i]=NULL;
  }


}

void leafset_insert(ChimeraState *state,ChimeraHost *host, int right_or_left, ChimeraHost **deleted, ChimeraHost **added){

  int i=0 ,size;
  ChimeraHost **p;
  ChimeraHost *tmp1,*tmp2;  
  ChimeraHost *input = host;
  Key dif1,dif2;
  RouteGlobal *routeglob = (RouteGlobal *) state->route;

  /*inert in right leafset*/
  if(right_or_left==1){
    size=leafset_size(routeglob->rightleafset);
    p=routeglob->rightleafset;
  }
  /*insert in left leafset*/
  else{
    size=leafset_size(routeglob->leftleafset);
    p=routeglob->leftleafset;
  }
   
  if(size==0) {
    p[0]=input;
    *added = input;
  }

  else {
    key_distance(state->log,&dif1,routeglob->me->key, p[i]->key);
    key_distance(state->log,&dif2,routeglob->me->key, host->key);
    while((i<size) && (key_comp(dif1,dif2)<=0)){
      if(key_equal(p[i]->key,input->key)){
	host_release(state, input);
	return;
      }
      i++;
      if(i<size){
	key_distance(state->log,&dif1,routeglob->me->key, p[i]->key);
	key_distance(state->log,&dif2,routeglob->me->key, host->key);
      }
    }
    
    tmp1=input;
    *added = input;
    
    while(i<LEAFSET_SIZE/2){
      tmp2=p[i];
      p[i++]=tmp1;
      tmp1=tmp2;
    } 

    /* there is leftover */
    if(size==LEAFSET_SIZE/2 && tmp2 != NULL)
      *deleted = tmp2;
  }
}

/** route_neighbors: 
** returns an array of #count# neighbor nodes with priority to closer nodes 
*/
ChimeraHost **route_neighbors(ChimeraState *state,int count){
  int i=0,j=0,Rsize=0,Lsize=0, index=0;
  int ret_size;
  ChimeraHost *tmp;
  ChimeraHost **hosts=(ChimeraHost **)malloc(sizeof(ChimeraHost *)*(LEAFSET_SIZE+1));
  ChimeraHost **ret=(ChimeraHost **)malloc(sizeof(ChimeraHost *)*(count+1));
  RouteGlobal *routeglob = (RouteGlobal *) state->route;

  pthread_mutex_lock(&routeglob->lock);

  Lsize=leafset_size(routeglob->leftleafset);
  Rsize=leafset_size(routeglob->rightleafset);

  if(count>Rsize+Lsize){
    ret_size=Rsize+Lsize;
  }
  else 
    ret_size=count;

  /* creat a jrb of leafset pointers sorted on distance */
  for(i=0;i<Lsize;i++) {
    tmp = routeglob->leftleafset[i];
    hosts[index++]=tmp;
  }
  
  for(i=0;i<Rsize;i++) {
    tmp = routeglob->rightleafset[i];
    hosts[index++]=tmp;
  }

  hosts[index]=NULL;
  /* sort aux */
  sort_hosts (state->log, hosts,routeglob->me->key,index);
  
  for(i=0;i<ret_size;i++) {
    tmp = hosts[i];
    ret[i] = host_get(state, tmp->name, tmp->port);
  }
  
  ret[i] = NULL;

  pthread_mutex_unlock(&routeglob->lock);
    
  return ret;
}


/** route_update:
** updated the routing table in regard to #host#. If the host is joining
** the network (and #joined# == 1), then it is added to the routing table
** if it is appropriate. If it is leaving the network (and #joined# == 0),
** then it is removed from the routing tables 
*/

void route_update(ChimeraState *state,ChimeraHost *host, int joined){
  int i,j;
  ChimeraHost *deleted = NULL, *added = NULL;
  RouteGlobal *routeglob = (RouteGlobal *)state->route;
  

  pthread_mutex_lock(&routeglob->lock);

  
  if(key_equal(routeglob->me->key, host->key)) {
    pthread_mutex_unlock(&routeglob->lock);
    return;
  }

  i = key_index(state->log,routeglob->me->key, host->key);
  j=hexalpha_to_int(host->key.keystr[i]);

  /*join*/
  if(joined) 
    {
      if(routeglob->table[i][j]!=NULL)
	host_release(state, routeglob->table[i][j]);
      
      routeglob->table[i][j]=host_get(state, host->name, host->port);
      leafset_update(state,host,joined,&deleted,&added);
    }
  
  /*delete*/
  else 
    {
      if(routeglob->table[i][j]!=NULL && key_equal(routeglob->table[i][j]->key,host->key))
	{
	  host_release(state, routeglob->table[i][j]);
	  routeglob->table[i][j]=NULL;
	}

      leafset_update(state,host,joined,&deleted,&added);
	
    }

  pthread_mutex_unlock(&routeglob->lock);

  if(deleted != NULL) {
    chimera_update_upcall(state, &(deleted->key),deleted,0);
  }
  if(added != NULL) {
    chimera_update_upcall(state, &(added->key),added,1);
  }
}

void gethostinfo(ChimeraHost *CHost){

  char tmp[64];
  CHost->name=(char *)malloc(sizeof(char)*64);
  printf("Enter key: ");
  scanf("%s", tmp);
  str_to_key(tmp, &(CHost->key));
  printf("Key :");
  key_print(CHost->key);
}

int hexalpha_to_int(int c)
{
  char hexalpha[] = "0123456789abcdef";
  int i;
  int answer = 0;

  for(i = 0; answer == 0 && hexalpha[i] != '\0'; i++)
  {
    if(hexalpha[i] == c)
    {
      answer = i ;
    }
  }

  return answer;
}

int power ( int base, int n){
  int i,p=1;

  for(i=1; i<=n; i++)
    p=p*base;
  return p;
}

void leafset_print(ChimeraState *state){
  int i;
  int Lsize,Rsize;
  RouteGlobal *routeglob = (RouteGlobal *) state->route;

  Lsize=leafset_size(routeglob->leftleafset);
  Rsize=leafset_size(routeglob->rightleafset);
  
  fprintf(stderr,"LEFT: ");
  for (i=0;i<Lsize;i++)
    fprintf(stderr,"%s ", routeglob->leftleafset[i]->key.keystr);
  fprintf(stderr,"\nRIGHT: ");
  for (i=0;i<Rsize;i++)
    fprintf(stderr,"%s ", routeglob->rightleafset[i]->key.keystr);
  fprintf(stderr,"\n");
}

void printTable(ChimeraState *state){

  int i,j;
  RouteGlobal *routeglob = (RouteGlobal *) state->route;

 /* print the table */

  fprintf(stderr,"------------------------------- TABLE-------------------------------\n");
  for ( i=0; i< MAX_ROW; i++){
    for(j=0;j<MAX_COL;j++){
      if( routeglob->table[i][j]!=NULL)
	fprintf(stderr,"%s ", routeglob->table[i][j]->key.keystr);
      else
	fprintf(stderr,"00000000 00000000 00000000 00000000  00000000");
    }
    fprintf(stderr,"\n");
  }
  fprintf(stderr,"----------------------------------------------------------------------\n");
}



